/*
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 * 
 * Contact Information:
 * Dambach-Werke GmbH
 * Elektronische Leitsysteme
 * Fritz-Minhardt-Str. 1
 * 76456 Kuppenheim
 * Phone: +49-7222-402-0
 * Fax: +49-7222-402-200
 * mailto: info@els.dambach.de
 */

package de.bsvrz.vew.engvew.engstelle;

import de.bsvrz.dav.daf.main.Data;
import de.bsvrz.dav.daf.main.DataDescription;
import de.bsvrz.dav.daf.main.ReceiveOptions;
import de.bsvrz.dav.daf.main.ReceiverRole;
import de.bsvrz.dav.daf.main.ResultData;
import de.bsvrz.dav.daf.main.config.Aspect;
import de.bsvrz.dav.daf.main.config.AttributeGroup;
import de.bsvrz.dav.daf.main.config.SystemObject;
import de.bsvrz.sys.funclib.debug.Debug;
import de.bsvrz.vew.engvew.verwaltung.VerwaltungEngstellen;

/**
 * Modelliert einen Unfall als spezielle Engstelle
 * 
 * @author Dambach Werke GmbH
 * @author Stefan Sans
 * @version $Revision: 1.3 $ / $Date: 2008/11/24 16:05:40 $ / ($Author: Sans $)
 */
public class Unfall extends Engstelle 
{
  /**
   * Debug-Logger für Logging-Ausgaben
   */
  private final static Debug _debug = Debug.getLogger();

  /**
   * Restkapazität in Fzg/h
   */
  private long _restKapazitaet = -1;

  /**
   * Erzeugt eine neue Instanz vom Typ Unfall
   *
   * @param verwaltung die Engstellenverwaltung, die für diese Instanz zuständig ist
   * @param davObjekt das Unfall-Systemobjekt
   */
  public Unfall(VerwaltungEngstellen verwaltung, SystemObject davObjekt)
  {
    super(verwaltung, davObjekt);

  }

  /**
   * @return Die Restkapazität dieses Unfalls oder -1,
   * wenn die Restkapazität nicht bestimmt werden konnte
   */
  public long getEngpassKapazitaet()
  {
    return _restKapazitaet;
  }

  /**
   * Bestimmt den Parameter WarteZeitMeldungNachRäumungUnfall der
   * Engstellenverwaltung
   *
   *  @return den Wert des Parameters
   */
  public long getWarteZeitMeldungNachRaeumungUnfall()
  {
    return _verwaltung.getWarteZeitMeldungNachRaeumungUnfall();
  }

  /*
   * (Kein Javadoc)
   * @see de.bsvrz.vew.vewallg.empf.EmpfaengerMitAnmeldung#anmelden()
   */
  public void anmelden()
  {
    super.anmelden();

    AttributeGroup atg = getConnection().getDataModel().getAttributeGroup("atg.unfallEigenschaften");
    Aspect asp = getConnection().getDataModel().getAspect("asp.parameterSoll");
    DataDescription datenbeschreibung = new DataDescription(atg, asp);

    getConnection().subscribeReceiver(this, getObjekt(), datenbeschreibung,
        ReceiveOptions.normal(), ReceiverRole.receiver());
  }

  /*
   * (Kein Javadoc)
   * @see de.bsvrz.vew.vewallg.empf.EmpfaengerMitAnAbmeldung#abmelden()
   */
  public void abmelden()
  {
    AttributeGroup atg = getConnection().getDataModel().getAttributeGroup("atg.unfallEigenschaften");
    Aspect asp = getConnection().getDataModel().getAspect("asp.parameterSoll");
    DataDescription datenbeschreibung = new DataDescription(atg, asp);

    getConnection().unsubscribeReceiver(this, getObjekt(), datenbeschreibung);

    super.abmelden();
  }

  /*
   * (Kein Javadoc)
   * @see de.bsvrz.dav.daf.main.ClientReceiverInterface#update(de.bsvrz.dav.daf.main.ResultData[])
   */
  public void update(ResultData[] resultData)
  {
    super.update(resultData);

    // Ausgaben
    for (ResultData data : resultData)
    {
      Aspect asp = data.getDataDescription().getAspect();
      AttributeGroup atg = data.getDataDescription().getAttributeGroup();
      short sv = data.getDataDescription().getSimulationVariant();

      _debug.finest("Daten empfangen mit " + atg.getPid() + "/" +
          asp.getPid() + "/" + sv  + ":\n" + data);

      if (asp.getPid().equals("asp.parameterSoll") &&
          atg.getPid().equals("atg.unfallEigenschaften"))
      {
        if (data.hasData())
        {
          _debug.finer("Bearbeite Daten mit " + atg.getPid() + "/" +
              asp.getPid() + "/" + sv + ":\n" + data);

          Data item = data.getData().getItem("RestKapazität");
          _restKapazitaet = item.asUnscaledValue().longValue();
          _debug.fine("Restkapazität ist " + _restKapazitaet + " für Unfall" + this);
        }
        else
        {
          _restKapazitaet = -1;
        }
        _debug.fine("Update Restkapazität für " + this + " Wert ist " + _restKapazitaet);
      }
    }
  }

  /**
   * Führt die Ablaufprüfung für diesen Unfall durch.
   * Die Zeit vom Ablauf der der (geschätzten) Dauer des Unfalls bis zum Prüfungszeitpunkt
   * wirde mit dem Parameter <code>WarteZeitMeldungNachRaeumungUnfall</code> verglichen.
   *
   * @return die Zeit in Minuten, die der Unfall abgelaufen ist, wenn die Zeit nach Unfallender
   * den Wert des Parmaters überschritten hat, 0 sonst.
   * Rückgabe von 0 bedeutet also positives, ingleich 0 negatives Ergebnis der Prüfung.
   * @throws IllegalStateException wenn die Prüfung nicht durchgeführt werden kann, aufgrund
   * etwa nichtversorgter Parameter, keine Gültigkeit spezifiziert etc.
   */
  public synchronized long pruefeAblauf()
    throws IllegalStateException
  {
    long warteZeitMeldungNachRaeumungUnfall = getWarteZeitMeldungNachRaeumungUnfall();
    if (warteZeitMeldungNachRaeumungUnfall < 0)
    {
      throw new IllegalStateException(this + " ist nicht prüfbar " +
        "da Parameter \"WarteZeitMeldungNachRaeumungUnfall\" nicht versorgt");
    }

    if (getGueltigkeit() == null)
    {
      throw new IllegalStateException(this + " ist nicht prüfbar " +
        "da Gültigkeit nicht spezifiziert");
    }

    long zeitpunkt = getConnection().getTime();

    if (getGueltigkeit().beginn <= zeitpunkt && // Engstelle beginnt nicht in der Zukunft
        getGueltigkeit().dauer > 0 && // Engstelle dauert nicht unendlich lang
        getGueltigkeit().beginn + getGueltigkeit().dauer + warteZeitMeldungNachRaeumungUnfall
        < zeitpunkt)
    {
      _debug.info(this + " ist seit mehr als " + warteZeitMeldungNachRaeumungUnfall / (60 * 1000) +
          "min. (" + ((zeitpunkt - (getGueltigkeit().beginn + getGueltigkeit().dauer)) /
          (60 * 1000)) + "min.) abgelaufen");
      return ((zeitpunkt - (getGueltigkeit().beginn + getGueltigkeit().dauer)) / (60 * 1000)) ;
    }

    ablaufMeldungen.setMeldungInaktiv(getIdAblaufMeldung());
    return 0;
  }

  /**
   * Instanz von <code>MeldungsBehandlung</code> zur Behandlung von Betriebsmeldungen zur
   * Ablaufprüfung dieses Unfalls
   */
  public final MeldungsBehandlung ablaufMeldungen = new MeldungsBehandlung();

  /**
   * Bildet eine Id zur Verwaltung von Betriebsmeldungen bezgl. der Ablaufprüfung
   * dieses Unfalls
   * 
   * @return die gebildete Id
   */
  public String getIdAblaufMeldung()
  {
    return getObjekt().getPid();
  }

}
